package maxmotion3d 
{
	import flash.display.BitmapData;
	import flash.display3D.*;
	import flash.display3D.textures.Texture;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.sampler.NewObjectSample;

	public class Model3D
	{
		protected var _vertices:Vector.<Vector.<Number>> = new Vector.<Vector.<Number>>();
		protected static const PER_VERTEX:int = 6; 
		protected var _indices:Vector.<Vector.<uint>> = new Vector.<Vector.<uint>>();
		protected var _vertexBuffer:Vector.<VertexBuffer3D> = new Vector.<VertexBuffer3D>();
		protected var _indexBuffer:Vector.<IndexBuffer3D> = new Vector.<IndexBuffer3D>(); 
		protected var _textures:Vector.<Texture> = new Vector.<Texture>();
		public var _pos:Vector3D = new Vector3D();
		public var _rotate:Vector3D = new Vector3D();
		public var _scale:Vector3D = new Vector3D(1,1,1);
		public var _animations:Vector.<Vector.<Vector.<Matrix3D>>>
			= new Vector.<Vector.<Vector.<Matrix3D>>>();
		protected var _time:int = 0;
		public var _currentAnimation:int = 0;
		public var _animationPlaying:Boolean = true;
		public var _loopAnimation:Boolean = true;
		public var _animationLast:Boolean = false;		
		protected var _current:int;
		public var _distance:Number;
		public var _intersectU:Number;
		public var _intersectV:Number;
		public var _visible:Boolean = true;
	
		public function Model3D(materialsNum:uint = 1) 
		{
			_vertices = new Vector.<Vector.<Number>>();
			for ( var i:int = 0; i < materialsNum; i++ )
			{
				_vertices.push(new Vector.<Number>());
			}
		}
		
		protected function init(context3D:Context3D,bitmapData:BitmapData,index:uint=0):void
		{
			if ( _vertices[index].length == 0 )
			{
				_vertexBuffer.length++;
				_indices.push(new Vector.<uint>());
				_indexBuffer.length++;
				_textures.length++;
				return;
			}
			var verticesNum:int = _vertices[index].length / PER_VERTEX;
			_vertexBuffer.push(context3D.createVertexBuffer(verticesNum, PER_VERTEX));
			_vertexBuffer[index].uploadFromVector(_vertices[index], 0, verticesNum);

			_indices.push(new Vector.<uint>());
			for ( var i:uint = 0; i < verticesNum; i++ )
			{
				_indices[index].push(i);
			}
			_indexBuffer.push(context3D.createIndexBuffer(verticesNum));
			_indexBuffer[index].uploadFromVector(_indices[index], 0, verticesNum);

			_textures.push(context3D.createTexture(
			bitmapData.width, bitmapData.height, Context3DTextureFormat.BGRA, false));
			_textures[index].uploadFromBitmapData(bitmapData);

			_currentAnimation = 0;
		}

		public function dispose():void
		{
			_vertices = new Vector.<Vector.<Number>>();
			_indices = new Vector.<Vector.<uint>>();
			_textures = new Vector.<Texture>();
			_vertexBuffer = new Vector.<VertexBuffer3D>();
			_indexBuffer = new Vector.<IndexBuffer3D>();
			_animations　= new Vector.<Vector.<Vector.<Matrix3D>>>();
		}
		
		protected function addVertex(x:Number, y:Number, z:Number,
		u:Number,v:Number,bone:Number,material:uint=0):void
		{
			_vertices[material].push(x,y,z,u,v,bone);
		}
		
		public function draw(scene:Scene3D):void
		{
			if ( !_visible )
			{
				return;
			}
			var bone:Matrix3D = new Matrix3D();
			for ( var i:int = 0; i < _textures.length; i++ )
			{
				if ( _vertices[i].length == 0 )
				{
					continue;
				}
				setVertices(scene._context3D,scene._cameraProj,scene._program,i);
				scene._context3D.setProgramConstantsFromMatrix(
				Context3DProgramType.VERTEX, 4, bone, true);
				scene._context3D.drawTriangles(_indexBuffer[i]);
			}
		}

		protected function setVertices(context3D:Context3D,camera:Matrix3D,program:Program3D,index:uint):void
		{
			// vertex position to attribute register 0
			context3D.setVertexBufferAt(0, _vertexBuffer[index], 0, Context3DVertexBufferFormat.FLOAT_3);
			// uv coordinates to attribute register 1
			context3D.setVertexBufferAt(1, _vertexBuffer[index], 3, Context3DVertexBufferFormat.FLOAT_2);
			// bone to attribute register 2
			context3D.setVertexBufferAt(2, _vertexBuffer[index], 5, Context3DVertexBufferFormat.FLOAT_1);
			// assign _texture to _texture sampler 0
			context3D.setTextureAt(0, _textures[index]);
			// assign shader program
			context3D.setProgram(program);
			
			var matrix:Matrix3D = new Matrix3D();
			matrix.prepend(camera);
			matrix.prependTranslation(_pos.x, _pos.y, _pos.z);
			matrix.prependRotation(_rotate.x, Vector3D.X_AXIS);
			matrix.prependRotation(_rotate.y, Vector3D.Y_AXIS);
			matrix.prependRotation(_rotate.z, Vector3D.Z_AXIS);
			matrix.prependScale(_scale.x, _scale.y, _scale.z);
			
			context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 0, matrix, true);
		}

		public function playAnimation(animation:int, loop:Boolean=true, last:Boolean=false):void
		{	
		}
		
		public function intersect(origin:Vector3D, dir:Vector3D):Boolean
		{
			for ( var i:int = 0; i < _vertices.length; i++ )
			{
				var v:Vector.<Number> = _vertices[i];
				for ( var j:int = 0; j < v.length; j += 18 )
				{
					var v0:Vector3D = new Vector3D(v[j     ], v[j + 1 ], v[j + 2 ]);
					var v1:Vector3D = new Vector3D(v[j + 6 ], v[j + 7 ], v[j + 8 ]);
					var v2:Vector3D = new Vector3D(v[j + 12], v[j + 13], v[j + 14]);
					if ( intersectTriangle(origin, dir, v0, v1, v2) )
					{
						return true;
					}
				}
			}
			return false;
		}

		protected function intersectTriangle(origin:Vector3D, dir:Vector3D,
			v0:Vector3D, v1:Vector3D, v2:Vector3D):Boolean
		{
			var e1:Vector3D = new Vector3D();
			var e2:Vector3D = new Vector3D();
			var pvec:Vector3D = new Vector3D();
			var tvec:Vector3D = new Vector3D();
			var qvec:Vector3D = new Vector3D();
			var det:Number;
			var t:Number;
			var u:Number;
			var v:Number;
			var inv_det:Number;

			e1 = v1.subtract(v0);
			e2 = v2.subtract(v0);

			pvec = dir.crossProduct(e2);
			det = e1.dotProduct(pvec);

			if (det > (1e-3))
			{
				tvec = origin.subtract(v0);
				u = tvec.dotProduct(pvec);
				if (u < 0.0 || u > det)
				{
					return false;
				}

				qvec = tvec.crossProduct(e1);

				v = dir.dotProduct(qvec);
				if (v < 0.0 || u + v > det)
				{
					return false;
				}
			}
			else if (det < -(1e-3))
			{
				tvec = origin.subtract(v0);

				u = tvec.dotProduct(pvec);
				if (u > 0.0 || u < det)
				{
					return false;
				}

				qvec = tvec.crossProduct(e1);

				v = dir.dotProduct(qvec);
				if (v > 0.0 || u + v < det)
				{
					return false;
				}
			}
			else
			{
				return false;
			}

			inv_det = 1.0 / det;

			t = e2.dotProduct(qvec);
			t *= inv_det;
			u *= inv_det;
			v *= inv_det;

			_distance = t;
			_intersectU = u;
			_intersectV = v;

			return true;
		}

		public function getPos():Vector3D
		{
			return new Vector3D(_pos.x,_pos.y,_pos.z);
		}

		public function forward(speed:Number):Vector3D
		{
			var pos:Vector3D = new Vector3D();
			var radian:Number = Math3D.toRadian(_rotate.y);
			
			pos.x = _pos.x + Math.sin(radian) * speed;
			pos.y = _pos.y;
			pos.z = _pos.z + Math.cos(radian) * speed;

			return pos;
		}

		public function dirRadian(dest:Vector3D):Number
		{
			var radian:Number;
			var dir:Vector3D = new Vector3D();

			dir = dest.subtract(_pos);

			if ( dir.x > 0.0 )
			{
				return -Math.atan(dir.z / dir.x) + Math.PI / 2;
			}
			else if ( dir.x < 0.0 )
			{
				return -Math.atan(dir.z / dir.x) - Math.PI / 2;
			}
			else if ( dir.z < 0 )
			{
				return Math.PI;
			}
			return 0;
		}

		public function dirDegree(dest:Vector3D):Number
		{
			var degree:Number = Math3D.toDegree(dirRadian(dest));

			if ( degree < 0 )
			{
				degree += 360;
			}
			if ( degree >= 360 )
			{
				degree -= 360;
			}
			return degree;
		}
		
		public function addRotateY(delta:Number):void
		{
			_rotate.y += delta;
			if ( _rotate.y < 0 )
			{
				_rotate.y += 360;
			}
			if ( _rotate.y >= 360 )
			{
				_rotate.y -= 360;
			}
		}
		
		public function posPlus(x:Number, y:Number, z:Number):Vector3D
		{
			var pos:Vector3D = new Vector3D();
			pos.x = _pos.x + x;
			pos.y = _pos.y + y;
			pos.z = _pos.z + z;
			return pos;
		}
	}
}
