package maxmotion3d 
{
	import com.adobe.utils.*;
	import flash.display.Stage;
	import flash.display.StageAlign;
	import flash.display.StageScaleMode;
	import flash.display3D.*;
	import flash.events.*;
	import flash.geom.Matrix3D;
	import flash.geom.Point;
	import flash.geom.Vector3D;
	import flash.display.Sprite;

	public class Scene3D
	{
		private var _sprite:Sprite;
		public var _context3D:Context3D;
		public  var _program:Program3D;		
		protected var _projection:PerspectiveMatrix3D = new PerspectiveMatrix3D();
		protected var _camera:PerspectiveMatrix3D = new PerspectiveMatrix3D();
		public var _up:Vector3D = new Vector3D(0, 1, 0);
		public var _fov:Number = 45;
		public var _aspect:Number;
		public var _zNear:Number = 1;
		public var _zFar:Number = 10000;
		public var _eye:Vector3D = new Vector3D(0, 0, 1000);
		public var _lookAt:Vector3D = new Vector3D(0, 0, 0);
		public var _cameraProj:Matrix3D = new Matrix3D();

		public function Scene3D(sprite:Sprite)
		{
			_sprite = sprite;
			_sprite.stage.stage3Ds[0].requestContext3D();
			_sprite.stage.scaleMode = StageScaleMode.NO_SCALE;
			_sprite.stage.align = StageAlign.TOP_LEFT;
		}
	
		public function initScene3D():void
		{
			_context3D = _sprite.stage.stage3Ds[0].context3D;			

			if ( !_context3D )
			{
				return;
			}

			var vertexShaderAssembler : AGALMiniAssembler = new AGALMiniAssembler();
			vertexShaderAssembler.assemble(Context3DProgramType.VERTEX,
				"m44 vt0, va0, vc[va2.x]\n" +
				"m44 op, vt0, vc0\n" +
				"mov v0, va1"
			);
			var fragmentShaderAssembler : AGALMiniAssembler= new AGALMiniAssembler();
			fragmentShaderAssembler.assemble(Context3DProgramType.FRAGMENT,
				"tex ft1, v0, fs0 <2d,repeat,linear,nomip>\n" +
				"mov oc, ft1"
			);

			_program = _context3D.createProgram();
			_program.upload(vertexShaderAssembler.agalcode, fragmentShaderAssembler.agalcode);
			_context3D.setCulling(Context3DTriangleFace.FRONT);
			_context3D.setBlendFactors( Context3DBlendFactor.SOURCE_ALPHA, Context3DBlendFactor.ONE_MINUS_SOURCE_ALPHA);
			
 			resizeHandler(null);
			_sprite.stage.addEventListener(Event.RESIZE, resizeHandler);
		}
		
        private function resizeHandler(event:Event):void
        {
			var w:int = _sprite.stage.stageWidth;
			var h:int = _sprite.stage.stageHeight;

			if ( w < 100 || h < 100 )
			{
				return;
			}
			
			_context3D.configureBackBuffer(w, h, 1, true);

			_aspect = (Number)(w)/(Number)(h);
			_projection.perspectiveFieldOfViewRH(Math3D.toRadian(_fov), _aspect, _zNear, _zFar);
		}
		
		public function setCamera():void
		{
			_camera.lookAtRH(_eye, _lookAt, _up);

			_cameraProj.identity();
			_cameraProj.prepend(_projection);
			_cameraProj.prepend(_camera);
		}
		
		public function getScreenPos(pos:Vector3D):Point
		{
			var v:Vector3D = _camera.transformVector(pos);
			var v2:Vector3D = _projection.transformVector(v);

			var half:Point = new Point(
				_sprite.stage.stageWidth  / 2,
				_sprite.stage.stageHeight / 2);

			return new Point(
				int(v2.x / v2.w * half.x) + half.x,
				-int(v2.y / v2.w * half.y) + half.y);
		}
		
		public function getEyePos():Vector3D
		{
			return new Vector3D(_eye.x,_eye.y,_eye.z);
		}
	}
}
